// rundir

// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Test that method expressions are correctly encoded
// in binary export data and can be imported again.
package ignore