// compiledir

// Copyright 2014 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Issue 7648: spurious "bad negated constant" for complex constants.

package ignored
