// errorcheckdir

// Copyright 2012 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// part one of issue 4124. Make sure that the compiler rejects access attempts.

package ignored

